import boto3, botocore, time
from enum import Enum

class AWSController:
    
    class AWSClients(Enum):
        EC2 = 'ec2'
        SSM = 'ssm'
    
    def __init__(self, aws_access_key, aws_secret_key, region_name = 'eu-central-1'):
        self.SYS_AWS_KEY = aws_access_key
        self.SYS_AWS_SECRET_KEY = aws_secret_key
        self.region_name = region_name
            
    def get_aws_client(self, region_name, client_type = AWSClients.EC2.value):
        """Returns an AWS account client for the specified region."""
        return boto3.client(
            client_type, 
            region_name=region_name,
            aws_access_key_id=self.SYS_AWS_KEY,
            aws_secret_access_key=self.SYS_AWS_SECRET_KEY
        )
    
    def getEC2Images(self, region_name):
        client = self.get_aws_client(region_name)
        return client.describe_images()
    
    def runSSMCommand(self, instance_id, command):
        """Runs an SSM command, waits for completion, and returns stdout, stderr, and exit code."""
        client = self.get_aws_client(self.region_name, self.AWSClients.SSM.value)

        # Define explicit exit codes for different failure scenarios
        exit_codes = {
            "Success": None,       # Use actual command exit code
            "Failed": 100,
            "Cancelled": 101,
            "TimedOut": 102,
            "Pending": 103,        # Shouldn't happen in final state
            "Delayed": 103         # Shouldn't happen in final state
        }

        # Send the command
        response = client.send_command(
            DocumentName="AWS-RunShellScript",
            Parameters={"commands": [command]},
            InstanceIds=[instance_id]
        )

        # Extract command ID
        command_id = response["Command"]["CommandId"]

        # Poll for command completion
        while True:
            result = client.get_command_invocation(CommandId=command_id, InstanceId=instance_id)
            status = result["Status"]

            if status in exit_codes:
                break  # Exit polling loop if final status reached

            time.sleep(2)  # Wait before polling again

        # Determine exit code
        exit_code = int(result.get("ResponseCode", -1)) if status == "Success" else exit_codes[status]

        # Return output as a dictionary
        return {
            "stdout": result.get("StandardOutputContent", ""),
            "stderr": result.get("StandardErrorContent", ""),
            "excode": exit_code,
        }

    
    def getEC2ImagesSizing(self, os_distributor, os_release, architecture='x86_64', free_tier_only=True):
        client = self.get_aws_client(self.region_name)
        filters = [
        {"Name": "state", "Values": ["available"]},  # Only available images
        ]

        images_response = client.describe_images(Filters=filters)
        os_distributor = os_distributor.lower()
        os_release = os_release.lower()
        architecture = architecture.lower()
    
        filtered_images = []
        
        for image in images_response["Images"]:
            name = image.get("Name", "").lower()
            image_arch = image.get("Architecture", "").lower()
    
            # Case-insensitive matching for OS distributor and release
            if os_distributor in name and os_release in name and architecture == image_arch:
                filtered_images.append(image)

        instance_types_response = client.describe_instance_types()
        free_tier_instances = {"t2.micro", "t3.micro"}
        instance_types = [
            {
                "type": instance["InstanceType"],
                "vcpus": instance["VCpuInfo"]["DefaultVCpus"],
                "memory_mb": instance["MemoryInfo"]["SizeInMiB"]
            }
            for instance in instance_types_response["InstanceTypes"]
            if not free_tier_only or instance["InstanceType"] in free_tier_instances
        ]
        return [
            {
                "image_id": image["ImageId"],
                "name": image.get("Name", "Unknown"),
                "platform": image.get("PlatformDetails", "Unknown"),
                "architecture": image.get("Architecture", "Unknown"),
                "available_sizing": instance_types if image["Architecture"] == architecture else []
            }
            for image in filtered_images
        ]
    
    def getEC2OSImage(self, os_distributor, os_release, architecture='x86_64'):
        client = self.get_aws_client(self.region_name)
        filters = [
            {"Name": "state", "Values": ["available"]},  # Only available AMIs
            {"Name": "architecture", "Values": [architecture]},  # x86_64 or arm64
            {"Name": "is-public", "Values": ["true"]}  # Exclude AWS Marketplace AMIs
       ]

        response = client.describe_images(Owners=["amazon"], Filters=filters)

        # Convert inputs to lowercase for case-insensitive matching
        os_distributor = os_distributor.lower()
        os_release = os_release.lower()

        filtered_images = [
            img for img in response["Images"]
            if os_distributor in img.get("Name", "").lower()
            and os_release in img.get("Name", "").lower()
        ]

        # Return the latest AMI by creation date
        if filtered_images:
            latest_ami = max(filtered_images, key=lambda x: x["CreationDate"])
            return latest_ami
        
        return {}

    def getVPCs(self, region_name):
        client = self.get_aws_client(region_name)
        return [
            {
                'vpc_id': v['VpcId'],
                'tags': v.get('Tags', [])
            }
            for v in client.describe_vpcs()['Vpcs']
        ]
    
    def getVPCSubnets(self, region_name):
        client = self.get_aws_client(region_name)
        return [
            {
                "id": sn['SubnetId'],
                "vpc": sn['VpcId'],
                "zone": sn['AvailabilityZone'].lower(),
                "cidr": sn['CidrBlock']
            }
            for sn in client.describe_subnets()['Subnets']
        ]
    
    def getAllRegions(self):
        client = self.get_aws_client(self.region_name)
        return [{"name": r['RegionName']} for r in client.describe_regions()['Regions']]
            
    def create_key_pair(self, instance_name):
        client = self.get_aws_client(self.region_name)
        # Generate a unique key pair name based on instance_name and current timestamp
        key_name = f"{instance_name}-{int(time.time())}"
        try:
            # Create a new key pair
            response = client.create_key_pair(KeyName=key_name)
            private_key = response['KeyMaterial']
            return private_key, key_name
        except client.exceptions.ClientError as e:
            if "already exists" in str(e):
                print(f"Key pair {key_name} already exists, using existing key.")
                return None, key_name
            else:
                raise    

    def createEC2Instance(self, regionName, vpcId, instanceType, operatingSystemId, operatingSystemArchitecture, osDiskType, osDiskSize, subnetId, isDryRun, securityGroupId=None, timeout = 300):
        """
        Creates an EC2 instance with a dynamically generated key pair.
        """
        client = self.get_aws_client(regionName)

        if not securityGroupId:
            security_groups = client.describe_security_groups(Filters=[{'Name': 'vpc-id', 'Values': [vpcId]}])['SecurityGroups']
            if security_groups:
                securityGroupId = security_groups[0]['GroupId']
            else:
                raise Exception("No security groups found in the VPC")

        keyName = f"{instanceType}-{int(time.time())}"  
        privateKey, keyName = self.create_key_pair(keyName)

        instance_data = client.run_instances(
            MaxCount=1,
            MinCount=1,
            BlockDeviceMappings=[{
                'Ebs': {
                    'VolumeSize': int(osDiskSize),
                    'VolumeType': osDiskType.lower(),
                    'Encrypted': False
                },
                'DeviceName': '/dev/sdh'
            }],
            ImageId=operatingSystemId,
            InstanceType=instanceType,
            KeyName=keyName,
            DryRun=isDryRun,
            NetworkInterfaces=[{
                'DeviceIndex': 0,
                'SubnetId': subnetId,
                'AssociatePublicIpAddress': True,
                'DeleteOnTermination': True,
                'Groups': [securityGroupId]
            }]
        )

        instanceId = instance_data['Instances'][0]['InstanceId']

        start_time = time.time()
        wait_time = 5
        publicIp = 'N/A'
    
        while time.time() - start_time < timeout:
            try:
                time.sleep(wait_time)
                wait_time = min(wait_time * 2, 30)
                response = client.describe_instances(InstanceIds=[instanceId])
                
                reservations = response.get("Reservations", [])
                
                if reservations:
                    instance = reservations[0]["Instances"][0]
                    publicIp = instance.get("PublicIpAddress", "N/A")
                    
                    if publicIp != "N/A":
                        break
          
            except botocore.exceptions.ClientError as e:
            
                error_code = e.response['Error']['Code']
                # that error is ok, instance could be in transition state
                if error_code == 'InvalidInstanceID.NotFound':
                    print(f"Instance {instanceId} are not initialized yet.\nWaiting for {wait_time} seconds before next try.")
                else:
                    raise
  
        if publicIp == "N/A":
            raise TimeoutError(f"Instance {instanceId} did not become public accessible within a {timeout} seconds.")

        key_pair = privateKey if privateKey else "Existing key used"

        ami_user_mapping = {
            "ami-0764af88874b6b852": "ec2-user",  # Amazon Linux
            "ami-0281e3f1612a178c5": "ec2-user",  # RHEL 9
            "ami-07eef52105e8a2059": "ubuntu",    # Ubuntu 24.04 LTS
            "ami-03b3b5f65db7e5c6f": "ubuntu",    # Ubuntu 22.04 LTS
            "ami-010ee62ec820032ce": "ec2-user",  # CentOS Stream 10
            "ami-011bb19bff62ccd36": "ec2-user"   # CentOS Stream 9
        }

        username = ami_user_mapping.get(operatingSystemId, "ec2-user")
        print(f"Instance {instanceId} created with key pair: {keyName}")
        print(f"Public IP: {publicIp}")
        print(f"Private Key:\n{key_pair}")
        print(f"Suggested SSH User: {username}")
        
        return {
            "InstanceId": instanceId,
            "PublicIP": publicIp,
            "KeyPair": key_pair,
            "SSHUser": username
        }

